import re

class Node:
    
    def __init__(self, surface, feature, stat):
        self.STAT_ID_FOR_NORMAL = 0
        self.STAT_ID_FOR_UNKNOWN = 1
        self.STAT_ID_FOR_BOS = 2
        self.STAT_ID_FOR_EOS = 3
        self.__surface = surface
        self.__feature = feature.split(',')
        self.__stat = stat

    def analyzable(self):
        return not self.bos() and not self.eos()

    def bos(self):
        return self.__stat == self.STAT_ID_FOR_BOS

    @property
    def conjugation1(self):
        return self.__feature[4]

    @property
    def conjugation2(self):
        return self.__feature[5]

    def eos(self):
        return self.__stat == self.STAT_ID_FOR_EOS

    def element_of_ikku(self):
        return self.normal()

    def first_of_ikku(self):
        if not self.first_of_phrase():
            return False
        if self.type == "記号" and not self.subtype1 in ["括弧開", "括弧閉"]:
            return False
        return True

    def first_of_phrase(self):
        if self.type in ["助詞", "助動詞"]:
            return False
        if self.subtype1 in ["非自立", "接尾"]:
            return False
        if self.subtype1 == "自立" and self.root_form in ["する", "できる"]:
            return False
        return True

    # def inspect
    #     to_s.inspect
    # end

    def last_of_ikku(self):
        if self.type in ["名詞接続", "格助詞", "係助詞", "連体化", "接続助詞", "並立助詞", "副詞化", "数接続", "連体詞"]:
            return False
        if self.conjugation2 == "連用タ接続":
            return False
        if self.conjugation1 == "サ変・スル" and self.conjugation2 == "連用形":
            return False
        if self.type == "動詞" and self.conjugation2 in ["仮定形", "未然形"]:
            return False
        if self.type == "名詞" and self.subtype1 == "非自立" and self.pronunciation == "ン":
            return False
        return True

    def last_of_phrase(self):
        return self.type != "接頭詞"

    def normal(self):
        return self.stat == self.STAT_ID_FOR_NORMAL

    @property
    def pronunciation(self):
        return self.__feature[8]

    def pronunciation_length(self):
        return len(self.pronunciation_mora())
        
    def pronunciation_mora(self):
        if not self.pronunciation == '':
            pronunciation = re.sub("[ぁ-ゔ]","[ァ-ヴ]",self.pronunciation)
            pronunciation = re.sub("[^アイウエオカ-モヤユヨラ-ロワヲンヴー]", "", pronunciation)
            return pronunciation
        else:
            return ''

    @property
    def root_form(self):
        return self.__feature[6]
 
    @property
    def stat(self):
        return self.__stat

    @property
    def subtype1(self):
        return self.__feature[1]

    @property
    def subtype2(self):
        return self.__feature[2]

    @property
    def subtype3(self):
        return self.__feature[3]

    @property
    def surface(self):
        return self.__surface

    @property
    def type(self):
        return self.__feature[0]



from .reviewer import *

__copyright__    = 'Copyright (C) 2020 Yohei Kirino'
__version__      = '0.1.1'
__license__      = 'MIT'
__author__       = 'Yohei Kirino'
__author_email__ = 'kirino.yohei@gmail.com'
__url__          = 'https://github.com/Gappo-Gappo/ikku'

__all__ = ['Reviewer']


import itertools

class BracketState:
    BRACKETS_TABLE = {
        "‘" : "’",
        "“" : "”",
        "（" : "）",
        "(" : ")",
        "［" : "］",
        "[" : "]",
        "{" : "}",
        "｛" : "｝",
        "〈" : "〉",
        "《" : "》",
        "「" : "」",
        "『" : "』",
        "【" : "】",
        "〔" : "〕",
        "<" : ">",
        "＜" : "＞",
    }

    def __init__(self):
        self.__stack = []
        self.__brackets_index = {v:True for v in list(itertools.chain.from_iterable(self.BRACKETS_TABLE.items()))}
        self.__inverted_brackets_table = {v:k for k,v in self.BRACKETS_TABLE.items()}

    def consume_all(self, surfaces):
        for surface in surfaces:
            self.__consume(surface)

    def odd(self):
        return not self.__even()

    def __consume(self, surface):
        if not self.__stack == [] and surface in self.__inverted_brackets_table and self.__inverted_brackets_table[surface] ==  self.__stack[-1]:
            self.__stack.pop()
        elif surface in self.__brackets_index:
            self.__stack.append(surface)

    def __even(self):
        return self.__stack == []



class Scanner:
    def __init__(self, exactly, nodes, rule):
        self.__exactly = exactly
        self.__nodes = nodes
        self.__rule = rule
        self.__count = 0
        self.__phrases = [[] for i in range(len(rule))]

    def scan(self):
        if self.has_valid_first_node():
            for node in self.__nodes:
                if self.consume(node):
                    if self.satisfied():
                        if not self.__exactly:
                            return self.__phrases
                else:
                    return
        if self.satisfied():
            return self.__phrases
        else:
            return

    def consume(self, node):
        if not node.element_of_ikku():
            return False
        if node.pronunciation_length() > self.max_consumable_length():
            return False
        if self.first_of_phrase() and not node.first_of_phrase():
            return False
        if node.pronunciation_length() == self.max_consumable_length() and not node.last_of_phrase():
            return False
        else:
            self.__phrases[self.phrase_index()].append(node)
            self.__count += node.pronunciation_length()
            return True

    def first_of_phrase(self):
        return self.__count in [sum(self.__rule[:i+1]) for i in range(len(self.__rule))]
    
    def has_full_count(self):
        return self.__count == sum(self.__rule)

    def has_valid_first_node(self):
        return self.__nodes[0].first_of_ikku()

    def has_valid_last_node(self):
        return self.__phrases[-1][-1].last_of_ikku()

    def max_consumable_length(self):
        return sum(self.__rule[:self.phrase_index()+1]) - self.__count

    def phrase_index(self):
        for i in range(len(self.__rule)):
            if self.__count < sum(self.__rule[:i+1]):
                return i
        return len(self.__rule) - 1

    def satisfied(self):
        return self.has_full_count() and self.has_valid_last_node()



import MeCab
from .node import Node
import os

class Parser:
    def parse(self, text):
        m = MeCab.Tagger()
        m.parse("")
        n = m.parseToNode(text)
        results = []
        while n:
            tmp = Node(n.surface, n.feature, n.stat)
            if tmp.analyzable():
                results.append(tmp)
            n = n.next 
        return results



from .parser import Parser
from .song import Song

class Reviewer:
    def __init__(self, rule = []):
        self.__parser = Parser()
        self.__rule = rule

    # Find one valid song from given text.
    # @return [Ikku::Song]
    def find(self, text):
        nodes = self.__parser.parse(text)
        for i in range(len(nodes)):
            song = Song(nodes[i:], exactly = False, rule = self.__rule)
            if song.valid():
                return song
        return None

    # Judge if given text is valid song or not.
    # @return [true, false]
    def judge(self, text):
        return Song(self.__parser.parse(text), exactly = True, rule = self.__rule).valid()

    # Search all valid songs from given text.
    # @return [Array<Array>]
    def search(self, text):
        nodes = self.__parser.parse(text)
        results = []
        for i in range(len(nodes)):
            song = Song(nodes[i:], exactly = False, rule = self.__rule)
            if song.valid():
                results.append(song)
        return results



# require "ikku/bracket_state"
from .scanner import Scanner
from .bracket_state import BracketState
import itertools

class Song:

    DEFAULT_RULE = [5, 7, 5]

    def __init__(self, nodes, exactly = False, rule = []):
        self.exactly = exactly
        self.__nodes = nodes
        self.__rule = rule or self.DEFAULT_RULE
        self.__phrases = self.scan()
        self.__bracket_state = BracketState()

    @property
    def phrases(self):
        results = []
        for n in range(len(self.__phrases)):
            results.append([s.surface for s in self.__phrases[n]])
        return results

    def valid(self):
        if not self.__phrases:
            return False
        elif self.has_odd_parentheses():
            return False
        else:
            return True

    def has_odd_parentheses(self):
        self.__bracket_state.consume_all(self.surfaces)
        return self.__bracket_state.odd()

    @property
    def nodes(self):
        return [s.surface for s in list(itertools.chain.from_iterable(self.__phrases))]

    @property
    def rule(self):
        return self.__rule

    def scan(self):
        return Scanner(self.exactly, self.__nodes, self.rule).scan()

    @property
    def surfaces(self):
        return self.nodes
